from numpy import *
from numpy.fft import *
from matplotlib import *
from scipy import *
#import elas
import sys
#import dbg

def homogenization(input):
    """FFTH homogenization
    INPUT = 
    input : dictionary consisting of following variables:
        L : ndarray of stiffness coefficients
        K0 : bulk modulus of reference stiffness
        G0 : shear modulus of reference stiffness
        Y : size of periodic unit cell
        tol : tolerance required for convergence
        maxit : maximal number of iterations
    OUTPUT = 
    L_homogenized : homogenized (effective) stiffness
    """
    d = ndim(input['L'][0][0])
    D = d*(d+1)/2
    N = array(shape(input['L'][0][0]))
    DN = hstack([D,N])
    L0 = get_stiff_Man(input['K0'], input['G0'])
    # calculation
    Gam = get_Gamma(N, d, D, input['Y'], input['K0'], input['G0'])
    solver = 11
    L_homogenized = zeros([D,D])#global macroscale stiffness - preallocation
    kit = zeros(D); resnorm = ones(D)
    for iL in arange(D):
        ep0 = zeros(D)
        ep0[iL] = 1
#        if solver == 1:
#            print(' ')
#            print('STANDARD ITERATIVE SCHEME with stress field')
#            ep = zeros(DN)
#            for i in arange(D):
#                ep[i]=ep0[i]
#            sig=mult_Lep(input['L'],ep,D,DN)
#            Fep=fft2(ep)
#            resnorm=1
#            kit=0
#            while (resnorm>tol) == (kit<maxit):
#                kit=kit+1
#                Fsig=fft2(sig)
#                Fep=Fep-elas.mult_GFsig(Gam,Fsig,D,DN)
#                ep_prev=ep
#                ep=real(ifft2(Fep))
#                sig=elas.mult_Lep(L,ep,D,DN)
#                dif=ep_prev-ep
#                rr=dot(dif.flatten(1),dif.flatten(1))
#                resnorm=sqrt(rr)
#        elif solver==11:
#            print(' ')
#            print('STANDARD with stress - CONJUGATE GRADIENT METHOD')
        X = zeros(DN)
        B = zeros(DN)        
        for i in arange(D):
            X[i] = ep0[i]
        AX = Ax_st(X,input['L'],L0,Gam,D,DN)
        R = B - AX
        P = R
        rr = dot(R.flatten(1),R.flatten(1))
        while (resnorm[iL]>input['tol']) == (kit[iL]<input['maxit']):
            kit[iL] = kit[iL]+1
            AP = Ax_st(P,input['L'],L0,Gam,D,DN)
            alp = rr/dot(P.flatten(1),AP.flatten(1))
            X = X+alp*P
            R = R-alp*AP
            rrnext = dot(R.flatten(1),R.flatten(1))
            bet = rrnext/rr
            rr = rrnext
            P = R+bet*P
            resnorm[iL] = sqrt(rr)
        ep = X
        sig = mult_Lep(input['L'],ep,D,DN)
        eqnorm = equinorm(sig,input['Y'],N,d)            
        for jL in arange(D):
            L_homogenized[jL][iL]=mean(sig[jL])
    return dict(L_homogenized=L_homogenized,kit=kit,resnorm=resnorm)

def get_stiff_Man(K,G):
    #returning stiffeness matrix with Mandel notation
    L=array([[K+4/3*G,K-2/3*G,0],[K-2/3*G,K+4/3*G,0],[0,0,G]])
    return L

def get_Gamma(N,d,D,Y,K0,G0):
    #function returning the discrete kernel Gamma
    xi=get_xi(Y,N,d)
    xi2=[]
    for i in arange(d):
        xi2.append(xi[i]**2)
    C=[1./(4*G0),(3.*K0+G0)/G0/(3*K0+4*G0)]
    Gamma=zeros([D,D])
    Gamma=Gamma.tolist()
    num11=transpose(tile(xi2[0],(N[1],1)))
    num22=tile(xi2[1],(N[0],1))
    num12=zeros(N)
    for i in arange(N[0]):
        for j in arange(N[1]):
            num12[i][j]=xi[0][i]*xi[1][j]
    denom2=num11+num22;denom2[0][0]=1
    denom4=denom2**2
    Gamma[0][0]=C[0]*4*num11/denom2-C[1]*num11**2/denom4
    Gamma[1][0]=Gamma[0][1]=-C[1]*num11*num22/denom4
    Gamma[2][0]=Gamma[0][2]=sqrt(2)*(C[0]*2*num12/denom2-C[1]*transpose(tile(xi[0]**3,(N[1],1)))*tile(xi[1],(N[0],1))/denom4)
    Gamma[1][1]=C[0]*4*num22/denom2-C[1]*tile(xi[1]**4,(N[0],1))/denom4
    Gamma[2][1]=Gamma[1][2]=sqrt(2)*(C[0]*2*num12/denom2-C[1]*transpose(tile(xi[0],(N[1],1)))*tile(xi[1]**3,(N[0],1))/denom4)
    Gamma[2][2]=2*(C[0]*(num11+num22)/denom2-C[1]*num11*num22/denom4)
    return Gamma

def get_KG_field(K_0,G_0,K_i,G_i,T):
    N=T.shape
    K=(ones(N)-T)*K_0+K_i*T
    G=(ones(N)-T)*G_0+G_i*T
    return [K,G]

def get_L(K,G):
    d=K.ndim
    D=d*(d+1)/2
    N=K.shape
    L=zeros([D,D])
    L=L.tolist()
    L[0][0]=K+4./3*G;
    L[1][0]=L[0][1]=K-2./3*G;
    L[1][1]=K+4./3*G;
    L[2][0]=L[0][2]=zeros(N);
    L[2][1]=L[1][2]=zeros(N);
    L[2][2]=2*G;
    return L

def mult_Lep(L,ep,D,DN):
    Lep=zeros(DN)
    for i in arange(D):
        for j in arange(D):
            Lep[i]=Lep[i]+L[i][j]*ep[j]
    return Lep

def mult_GFsig(Gam,Fsig,D,DN):
    GFsig=zeros(DN,dtype=complex128)
    for i in arange(D):
        GFsig[i]=Gam[i][0]*Fsig[0]+Gam[i][1]*Fsig[1]+Gam[i][2]*Fsig[2]
    return GFsig

def converse_moduli(valA,valB,RvalA,RvalB):
    pass
    return [RvalA,RvalB]

def Ax_st(ep,L,L0,Gam,D,DN):
    sig=mult_Lep(L,ep,D,DN)
    Fep=mult_GFsig(Gam,fft2(sig),D,DN)
    AX=real(ifft2(Fep));
    return AX

def Ax_pol(ep,L,L0,Gam,D,DN): #under construction
    tau=zeros(DN)
    for i in arange(D):
        for j in arange(D):
            tau[i]=tau[i]+(L[i][j]-L0[i][j])*ep[j]
    Fep=mult_GFsig(Gam,fft2(tau),D,DN)
    AX=ep+real(ifft2(Fep));
    return AX

def get_xi(Y,N,d):
    ind_mean=ceil(N/2.-1)
    xi=[]
    for i in arange(d):
        ZN=arange(fix(.5-N[i]/2.),fix(N[i]/2.+1))
        ZNshift=hstack((ZN[ind_mean[i]:N[i]],ZN[0:ind_mean[i]]))
        xi.append(ZNshift/Y[i])
    return xi

def get_xiM(Y,N,d):
    xi=get_xi(Y,N,d)
    xiM=zeros(hstack([d,N]))
    xiM[0]=transpose(tile(xi[0],(N[1],1)))
    xiM[1]=tile(xi[1],(N[0],1))
    return xiM

def equinorm(sig,Y,N,d):
    xiM=get_xiM(Y,N,d)
    sig[2]=sig[2]/sqrt(2)#unMandel notation
    Fsig=fft2(sig)
    Fsig0=real([[Fsig[0,0,0],Fsig[2,0,0]],[Fsig[2,0,0],Fsig[1,0,0]]])
    #Fsig0=Fsig0/(N[0]*N[1])
    R=zeros(hstack([d,N]),dtype=complex128)
    R[0]=xiM[0]*Fsig[0]+xiM[1]*Fsig[2]
    R[1]=xiM[0]*Fsig[2]+xiM[1]*Fsig[1]
    Rsum=R[0]+R[1]
    eqnorm=linalg.norm(Rsum.flatten(1))/linalg.norm(Fsig0.flatten(1))
    return eqnorm

def get_K0_G0(K,G):
    K0 = (K.max() + K.min())/2
    G0 = (G.max() + G.min())/2
    return K0,G0
    
def get_L0(L,D):
    L0 = zeros([D,D])
    for i in arange(D):
        L0[i][i] = max(L)
    return L0
    
    
